/*	alg.c
 *
 *	Detect changes in a video stream.
 *	Copyright 2001 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include "motion.h"
#include "video.h"
#include "conf.h"
#include "alg.h"
#include "event.h"

/* locate the center of the movement. */
struct coord alg_locate_center (struct images *imgs, int width, int height)
{
	char *out=imgs->out;
	int x, y, centc=0;
	struct coord cent;

	cent.x=0;
	cent.y=0;
	/* Locate movement */
	for (y=0; y<height; y++) {
		for (x=0; x<width; x++) {
			if (*(out++)) {
				cent.x+=x;
				cent.y+=y;
				centc++;
			}
		}
	}
	if (centc) {
		cent.x=cent.x/centc;
		cent.y=cent.y/centc;
	}
	return (cent);
}

/* draw a box around the movement */
void alg_locate (struct coord *cent, struct images *imgs, int width, int height)
{
	char *out=imgs->out, *new=imgs->new;
	int x, y, maxx=0, maxy=0, minx=width, miny=height;
	int centc=0, xdist=0, ydist=0;

	centc=0;
	for (y=0; y<height; y++) {
		for (x=0; x<width; x++) {
			if (*(out++)) {
				if (x > cent->x) xdist+=x-cent->x;
				if (y > cent->y) ydist+=y-cent->y;
				if (x < cent->x) xdist+=cent->x-x;
				if (y < cent->y) ydist+=cent->y-y;
				centc++;
			}
		}	
	}
	if (centc) {
		minx=cent->x-xdist/centc*2;
		maxx=cent->x+xdist/centc*2;
		/* Make the box a litle bigger in y direction to make sure the
		 * heads fit in */
		miny=cent->y-ydist/centc*3;
		maxy=cent->y+ydist/centc*2;
	}
	if (minx < 0) minx=0;
	if (miny < 0) miny=0;
	if (maxx > width-1) maxx=width-1;
	if (maxy > height-1) maxy=height-1;
	if (maxx < 0) maxx=0;
	if (maxy < 0) maxy=0;
	if (minx > width-1) minx=width-1;
	if (miny > height-1) miny=height-1;
	cent->width=maxx-minx;
	cent->height=maxy-miny;
	out=imgs->out;
	
	/* Draw a box around the movement */
	for (x=minx; x<=maxx; x++) {
		new[x+width*miny]=~new[x+width*miny];
		new[x+width*maxy]=~new[x+width*maxy];
		out[x+width*miny]=~out[x+width*miny];
		out[x+width*maxy]=~out[x+width*maxy];
	}
	for (y=miny; y<=maxy; y++) {
		new[minx+y*width]=~new[minx+y*width];
		new[maxx+y*width]=~new[maxx+y*width];
		out[minx+y*width]=~out[minx+y*width];
		out[maxx+y*width]=~out[maxx+y*width];
	}
}

#define NORM			100
#define ABS(x)			((x)<0 ? -(x) : (x))
#define DIFF(x, y)		(ABS((x)-(y)))
#define NDIFF(x, y)		(ABS(x)*NORM/(ABS(x)+2*DIFF(x,y)))
#define NRES(in, min, max)	((in)<(min)?NDIFF(min,in):((in)>(max)?NDIFF(max,in):100))

int alg_predict_eval_set(struct predict *min, struct predict *max, struct predict *in)
{
	return (
		NRES(in->x, min->x, max->x)+
		NRES(in->y, min->y, max->y)+
		NRES(in->dx, min->dx, max->dx)+
		NRES(in->dy, min->dy, max->dy)+
		NRES(in->width, min->width, max->height)+
		NRES(in->height, min->height, max->height)+
		NRES(in->diffs, min->diffs, max->diffs)+
		NRES(in->length, min->length, max->length)
	)/8;
}

void alg_predict_eval(struct context *cnt)
{
	struct predict_set *set=cnt->predict_sets;
	int high=0, i;
	struct predict_set *highset=set;

	while (set) {
		i=alg_predict_eval_set(&set->min, &set->max, &cnt->predict);
		printf("%s: %d\n", set->name, i);
		if (i>high) {
			high=i;
			highset=set;
		}
		set=set->next;
	}
	if (highset && high>=cnt->conf.predict_threshold) {
		printf("Best guess: %s (%d%%)\n", highset->name, high);
	} else {
		printf("Unknown type of movement\n");
	}
}

void alg_predict(struct coord *box, struct context *cnt, int diffs, int first)
{
	struct predict *avg=&cnt->predict;
	struct coord *lastbox=&cnt->predict_lastbox;

	if (first) {
		memcpy(&cnt->predict_lastbox, box, sizeof(struct coord));
		cnt->last_diffs=diffs;
		memset(avg, 0, sizeof(struct predict));
		avg->length=1;
		avg->x=box->x;
		avg->y=box->y;
		avg->width=box->width;
		avg->height=box->height;
		avg->diffs=diffs;
		return;
	}
	avg->length++;
	avg->x+=box->x;
	avg->y+=box->y;
	avg->width+=box->width;
	avg->height+=box->height;
	avg->dx+=box->x-lastbox->x;
	avg->dy+=box->y-lastbox->y;
	avg->diffs+=diffs;

	memcpy(&cnt->predict_lastbox, box, sizeof(struct coord));
	cnt->last_diffs=diffs;
	return;
}

void alg_predict_result(struct context *cnt, char *filename)
{
	struct predict *avg=&cnt->predict;
	FILE *fddesc;

	if (!avg->length)
		return;
	avg->x/=avg->length;
	avg->y/=avg->length;
	avg->width/=avg->length;
	avg->height/=avg->length;
	avg->diffs/=avg->length;
	printf("length: %d\t", avg->length);
	printf("x: %d\ty: %d\twidth: %d\theight: %d\tdiffs: %d\t", avg->x, avg->y, avg->width, avg->height, avg->diffs);
	printf("dx:%d\tdy:%d\n", avg->dx, avg->dy);
	alg_predict_eval(cnt);
	if ((fddesc=fopen(filename, "w"))==NULL) {
		perror("Error writing motion description file");
		return;
	}
	fprintf(fddesc, "Auto generated description\n");
	fprintf(fddesc, "%d\t", avg->length);
	fprintf(fddesc, "%d\t%d\t%d\t%d\t%d\t", avg->x, avg->y, avg->width, avg->height, avg->diffs);
	fprintf(fddesc, "%d\t%d\n", avg->dx, avg->dy);
	fprintf(fddesc, "%d\t", avg->length);
	fprintf(fddesc, "%d\t%d\t%d\t%d\t%d\t", avg->x, avg->y, avg->width, avg->height, avg->diffs);
	fprintf(fddesc, "%d\t%d\n", avg->dx, avg->dy);
	fclose(fddesc);
}


void alg_noise_tune (struct context *cnt, int motion)
{
	struct images *imgs=&cnt->imgs;
	int i;
	char *ref=imgs->ref;
	char *new=imgs->new;
	int diff, sum=0, count=0;
	unsigned char *mask=imgs->mask;

	i=imgs->motionsize;
	for (; i>0; i--) {
		diff=*ref-*new;
		if (mask)
			diff=((diff**mask++)/255);
		sum+=ABS(diff);
		if (diff)
			count++;
		ref++;
		new++;
	}
	sum/=count/4;
	if (!motion || sum>4*cnt->noise)
		cnt->noise=(cnt->noise+sum)/2;
	printf("New noise level: %d\n", cnt->noise);
}

void alg_threshold_tune (struct context *cnt, int diffs, int motion)
{
	int i;
	int sum=0, top=diffs;

	if (!diffs)
		return;

	printf("%d\n", diffs);
	if (motion)
		diffs=cnt->threshold/4;

	for (i=0; i<THRESHOLD_TUNE_LENGTH-1; i++)
	{
		sum+=cnt->diffs_last[i];
		if (cnt->diffs_last[i+1] && !motion)
			cnt->diffs_last[i]=cnt->diffs_last[i+1];
		else
			cnt->diffs_last[i]=cnt->threshold/4;
		if (cnt->diffs_last[i]>top)
			top=cnt->diffs_last[i];
	}
	sum+=cnt->diffs_last[i];
	cnt->diffs_last[i]=diffs;

	sum/=THRESHOLD_TUNE_LENGTH/4;
	if (sum<top*2)
		sum=top*2;
	if (sum < cnt->conf.max_changes)
			cnt->threshold=(cnt->threshold+sum)/2;
	printf("New threshold: %d\n", cnt->threshold);
}

/*	The heart of the motion detector
 *	Basicly: ref-new and count the pixels that are over a noise threshold
 */
int alg_diff_standard (struct context *cnt)
{
	struct images *imgs=&cnt->imgs;
	int i, diffs=0;
	long int level=0;
	int noise=cnt->noise;
	char *ref=imgs->ref;
	char *new=imgs->new;
	char *out=imgs->out;
	unsigned char *mask=imgs->mask;


	/* If the average level of the picture is to low, compensate by 
	 * lowering the noise threshold
	 */
	if (cnt->conf.nightcomp) {
		i=imgs->motionsize;
		for (i--; i>=0; i--) {
			level+=(unsigned char)new[i];
		}
		level/=imgs->motionsize;
		if (level < noise*2)
			noise*=2;
	}

	i=imgs->motionsize;
	for (; i>0; i--) {
		*out=*ref-*new;
		if (mask)
			*out=((int)((char)*out**mask++)/255);
		if (*out >  noise ||
		    *out < -noise ) {
			if (!cnt->conf.realmotion) *out=*new;
			diffs++;
		} else {
			if (!cnt->conf.realmotion) *out=0;
		}
		out++;
		ref++;
		new++;
	}

	return diffs;
}

/*
	Very fast diff function, does not do nightcompensation or mask
	overlaying.
*/
int alg_diff_fast (struct context *cnt)
{
	struct images *imgs=&cnt->imgs;
	int i, diffs=0;
	int noise=cnt->noise;
	char *out=imgs->out;
	char *ref=imgs->ref;
	char *new=imgs->new;

	memset(out, 0, imgs->size);
	i=imgs->motionsize;
	for (; i>0; i-=40) {
		*out=*ref-*new;
		if (*out >  noise ||
		    *out < -noise ) {
			if (!cnt->conf.realmotion) *out=*new;
			diffs++;
		} else {
			if (!cnt->conf.realmotion) *out=0;
		}
		out+=40;
		ref+=40;
		new+=40;
	}
	diffs=diffs*40;
	return diffs;
}

/*
	diff_hybrid uses diff_fast to quickly decide if there is anything worth
	sending to diff_standard.
*/
int alg_diff_hybrid (struct context *cnt)
{
	int diffs;
	
	diffs=alg_diff_fast(cnt);
	if (diffs > cnt->conf.max_changes/2)
		diffs=alg_diff_standard(cnt);
	return diffs;
}

int alg_diff (struct context *cnt)
{
	return alg_diff_hybrid(cnt);
}


/* Detect a sudden massive change in the picture.
   It is assumed to be the light being switched on or a camera displacement.
   In any way the user doesn't think it is worth capturing.
 */
int alg_lightswitch (struct context *cnt, int diffs, int dev, int pipe, int mpipe)
{
	struct images *imgs=&cnt->imgs;
	int i, j;
	
	/* is 2/3 of the image right?  */
	if (diffs > imgs->width*imgs->height*2) {
		printf("lightswitch detected\n");
		/* wait for the device to settle and update the
		 * reference frame to prevent the next frames
		 * from being detected
		 * if 1/3 of the frame is still moving asume that the camera
		 * has not yet compensated.
		 */
		i=10;
		while(i && (imgs->new=vid_next(cnt, dev, imgs->new, imgs->width, imgs->height))) {
			if (alg_diff(cnt)
			    > imgs->width*imgs->height)
				i=10;
			j=imgs->width*imgs->height;
			for (j--; j>=0; j--) {
				imgs->ref[j]=imgs->new[j];
			}
			event(EVENT_IMAGE, cnt, imgs->new, &pipe, cnt->currenttime);
			event(EVENT_IMAGEM, cnt, imgs->out, &mpipe, cnt->currenttime);
			i--;
		}
		return 0;
	}
	return diffs;
}

int alg_switchfilter (struct context *cnt, int diffs, struct coord *cent)
{
	int linediff=diffs/cnt->conf.height;
	struct images *imgs=&cnt->imgs;
	char *out=cnt->imgs.out;
	int y, x, line;
	int lines=0, vertlines=0;
	char tmp[80];

	for (y=0; y<cnt->conf.height; y++) {
		line=0;
		for (x=0; x<cnt->conf.width; x++) {
			if (*(out++)) {
				line++;
			}
		}
		if (line>cnt->conf.width/18) {
			vertlines++;
		}
		if (line>linediff*2) {
			lines++;
		}
	}
//	printf("%d %d %d %d  %d\n", lines, vertlines, linediff, lines-vertlines, diffs);
	if (vertlines>cnt->conf.height/10 && lines<vertlines/3 && 
	    (vertlines>cnt->conf.height/4 || lines-vertlines>lines/2)) {
	    	sprintf(tmp, "%d %d", lines, vertlines);
	    	draw_text(cnt->imgs.new, 10, 10, 288, 352, tmp, cnt->imgs.type);
		return diffs;
	}
	return 0;
}
